@@ -1,42 +1,69 @@
 import yaml
 import io
 from os import path
+import warnings
 
 from snsql.sql.reader.base import NameCompare
 
-# implements spec at https://docs.google.com/document/d/1Q4lUKyEu2W9qQKq6A0dbo0dohgSUxitbdGhX97sUNOM/
-
+# implements spec at https://docs.smartnoise.org/en/stable/sql/metadata.html
 
 class Metadata:
     """Information about a collection of tabular data sources"""
 
-    def __init__(self, tables, engine, compare=None):
+    def __init__(self, tables, engine=None, compare=None, dbname=None):
         """Instantiate a metadata object with information about tabular data sources
 
         :param tables: A list of Table descriptions
         :param engine: The name of the database engine used to query these tables.  Used for engine-
             specific name escaping and comparison.  Set to None to use default semantics.
+        :param compare: A NameCompare object used to compare table names.  Set to None to use comparison rules
+            associated with engine.
+        :param dbname: The name of the database.  Used to match 3-part object names like dbname.schema.table.  Set to None to match any database name.
         """
         self.m_tables = dict([(t.table_name(), t) for t in tables])
         self.engine = engine if engine is not None else "Unknown"
         self.compare = NameCompare.get_name_compare(engine) if compare is None else compare
+        self.dbname = dbname if dbname else None
 
     def __getitem__(self, tablename):
         schema_name = ""
+        dbname = ""
         parts = tablename.split(".")
-        if len(parts) == 2:
+        if len(parts) == 3:
+            dbname, schema_name, tablename = parts
+        elif len(parts) == 2:
             schema_name, tablename = parts
         for tname in self.m_tables.keys():
             table = self.m_tables[tname]
-            if self.compare.schema_match(
-                schema_name, table.schema
-            ) and self.compare.identifier_match(tablename, table.name):
-                table.compare = self.compare
-                return table
+
+            # check if table name matches
+            if not self.compare.identifier_match(tablename, table.name):
+                # should this really check tablename?  or just the final part?
+                continue
+
+            # check if schema name matches
+            if not schema_name:
+                if table.schema:
+                    if not self.compare.schema_match(schema_name, table.schema):
+                        continue
+                else:
+                    pass
+            else:
+                if not self.compare.schema_match(schema_name, table.schema):
+                    continue
+
+            # check if dbname matches
+            if dbname and self.dbname and not self.compare.identifier_match(dbname, self.dbname):
+                continue
+
+            # all check passed, return table
+            table.compare = self.compare
+            return table
         return None
 
     def __str__(self):
-        return "\n\n".join([str(self.m_tables[table]) for table in self.m_tables.keys()])
+        desc = f"{self.dbname if self.dbname else '(no database name)'} using engine {self.engine}\n"
+        return  desc + "\n".join([str(self.m_tables[table]) for table in self.m_tables.keys()])
 
     def tables(self):
         return [self.m_tables[tname] for tname in self.m_tables.keys()]
@@ -60,12 +87,8 @@ class Metadata:
     def from_(cls, val):
         if isinstance(val, Metadata):
             return val
-        elif isinstance(val, str):
-            if path.exists(val):
-                return cls.from_file(val)
-            else:
-                # could allow YAML strings here.  Throw for now
-                raise ValueError(f"Unable to load metadata path {val}")
+        elif isinstance(val, (str, io.IOBase)):
+            return cls.from_file(val)
         elif isinstance(val, dict):
             return cls.from_dict(val)
         else:
@@ -80,15 +103,14 @@ class Metadata:
     Common attributes for a table or a view
 """
 
-
 class Table:
     """Information about a single tabular data source"""
-
     def __init__(
         self,
         schema,
         name,
         columns,
+        *ignore,
         rowcount=0,
         rows_exact=None,
         row_privacy=False,
@@ -123,6 +145,15 @@ class Table:
         self.m_columns = dict([(c.name, c) for c in columns])
         self.compare = None
 
+        if clamp_columns:
+            for col in self.m_columns.values():
+                if col.typename() in ["int", "float"] and (col.lower is None or col.upper is None):
+                    if col.sensitivity is not None:
+                        raise ValueError(
+                            f"Column {col.name} has sensitivity and no bounds, but table specifies clamp_columns. "
+                            "clamp_columns should be False, or bounds should be provided."
+                        )
+
     def __getitem__(self, colname):
         for cname in self.m_columns.keys():
             col = self.m_columns[cname]
@@ -135,13 +166,14 @@ class Table:
 
     def __str__(self):
         return (
-            str(self.schema)
-            + "."
+            "  "
+            + str(self.schema)
+            + "." if self.schema else ""
             + str(self.name)
             + " ["
             + str(self.rowcount)
-            + " rows]\n\t"
-            + "\n\t".join([str(self.m_columns[col]) for col in self.m_columns.keys()])
+            + " rows]\n    "
+            + "\n    ".join([str(self.m_columns[col]) for col in self.m_columns.keys()])
         )
 
     def key_cols(self):
@@ -156,108 +188,172 @@ class Table:
 
     def __iter__(self):
         return self.columns()
-
     def table_name(self):
         return (self.schema + "." if len(self.schema.strip()) > 0 else "") + self.name
 
-
 class String:
     """A column with string data"""
-
-    def __init__(self, name, card, is_key=False, bounded=False):
+    def __init__(
+        self, 
+        name, 
+        *ignore, 
+        card=None, 
+        is_key=False,
+        nullable=True, 
+        missing_value=None
+    ):
         self.name = name
         self.card = card
         self.is_key = is_key
-        self.bounded = bounded
-
+        self.nullable = nullable
+        self.missing_value = missing_value
+        if self.missing_value is not None:
+            self.nullable = False
     def __str__(self):
         return ("*" if self.is_key else "") + str(self.name) + " (card: " + str(self.card) + ")"
-
     def typename(self):
         return "string"
-
+    @property
+    def unbounded(self):
+        return self.card is None
 
 class Boolean:
     """A column with True/False data"""
-
-    def __init__(self, name, is_key=False, bounded=False):
+    def __init__(
+        self, 
+        name, 
+        *ignore, 
+        is_key=False,
+        nullable=True, 
+        missing_value=None
+    ):
         self.name = name
         self.is_key = is_key
-        self.bounded = bounded
-
+        self.nullable = nullable
+        self.missing_value = missing_value
+        if self.missing_value is not None:
+            self.nullable = False
     def __str__(self):
         return ("*" if self.is_key else "") + str(self.name) + " (boolean)"
-
     def typename(self):
         return "boolean"
-
+    @property
+    def unbounded(self):
+        return False
 
 class DateTime:
     """A date/time column"""
-
-    def __init__(self, name, is_key=False, bounded=False):
+    def __init__(
+        self, 
+        name, 
+        *ignore, 
+        is_key=False,
+        nullable=True, 
+        missing_value=None
+    ):
         self.name = name
         self.is_key = is_key
-        self.bounded = bounded
-
+        self.nullable = nullable
+        self.missing_value = missing_value
+        if self.missing_value is not None:
+            self.nullable = False
     def __str__(self):
         return ("*" if self.is_key else "") + str(self.name) + " (datetime)"
-
     def typename(self):
         return "datetime"
-
+    @property
+    def unbounded(self):
+        return True
 
 class Int:
     """A column with integer data"""
-
-    def __init__(self, name, minval=None, maxval=None, is_key=False, bounded=False):
+    def __init__(
+        self, 
+        name, 
+        *ignore, 
+        lower=None, 
+        upper=None,
+        is_key=False,
+        sensitivity=None, 
+        nullable=True, 
+        missing_value=None
+    ):
         self.name = name
-        self.minval = minval
-        self.maxval = maxval
+        self.lower = lower
+        self.upper = upper
         self.is_key = is_key
-        self.unbounded = minval is None or maxval is None
-        self.bounded = bounded
+        self.sensitivity = sensitivity
+        self.nullable = nullable
+        self.missing_value = missing_value
+        if self.missing_value is not None:
+            self.nullable = False
+
+        if self.lower is not None and self.upper is not None and self.sensitivity is not None:
+            sens_a = max(abs(self.lower), abs(self.upper))
+            sens_b = abs(self.upper - self.lower)
+            if self.sensitivity != sens_a and self.sensitivity != sens_b:
+                warnings.warn(
+                    f"Sensitivity for {self.name} will override sensitivity derived from bounds"
+                )
 
     def __str__(self):
-        bounds = "unbounded" if self.unbounded else str(self.minval) + "," + str(self.maxval)
+        bounds = "unbounded" if self.unbounded else str(self.lower) + "," + str(self.upper)
         return ("*" if self.is_key else "") + str(self.name) + " [int] (" + bounds + ")"
-
     def typename(self):
         return "int"
-
+    @property
+    def unbounded(self):
+        return self.lower is None or self.upper is None
 
 class Float:
     """A floating point column"""
-
-    def __init__(self, name, minval=None, maxval=None, is_key=False, bounded=False):
+    def __init__(
+        self, 
+        name, 
+        *ignore,
+        lower=None, 
+        upper=None, 
+        is_key=False,
+        sensitivity=None, 
+        nullable=True, 
+        missing_value=None
+    ):
         self.name = name
-        self.minval = minval
-        self.maxval = maxval
+        self.lower = lower
+        self.upper = upper
         self.is_key = is_key
-        self.unbounded = minval is None or maxval is None
-        self.bounded = bounded
+        self.sensitivity = sensitivity
+        self.nullable = nullable
+        self.missing_value = missing_value
+
+        if self.lower is not None and self.upper is not None and self.sensitivity is not None:
+            sens_a = max(abs(self.lower), abs(self.upper))
+            sens_b = abs(self.upper - self.lower)
+            if self.sensitivity != sens_a and self.sensitivity != sens_b:
+                warnings.warn(
+                    f"Sensitivity for {self.name} will override sensitivity derived from bounds"
+                )
 
     def __str__(self):
-        bounds = "unbounded" if self.unbounded else str(self.minval) + "," + str(self.maxval)
+        bounds = "unbounded" if self.unbounded else str(self.lower) + "," + str(self.upper)
         return ("*" if self.is_key else "") + str(self.name) + " [float] (" + bounds + ")"
-
     def typename(self):
         return "float"
-
+    @property
+    def unbounded(self):
+        return self.lower is None or self.upper is None
 
 class Unknown:
     """Column is unknown type.  Will be ignored.  May not be used in queries."""
-
     def __init__(self, name):
         self.name = name
         self.is_key = False
-
     def __str__(self):
         return str(self.name) + " (unknown)"
-
     def typename(self):
         return "unknown"
-
+    def unbounded(self):
+        return True
 
 class CollectionYamlLoader:
     def __init__(self, file):
@@ -271,6 +367,8 @@ class CollectionYamlLoader:
                 raise
             return self._create_metadata_object(c_s)
         else:
+            if not path.exists(self.file):
+                raise ValueError(f"Unable to load metadata path {self.file}")
             with open(self.file, "r") as stream:
                 try:
                     c_s = yaml.safe_load(stream)
@@ -279,6 +377,8 @@ class CollectionYamlLoader:
             return self._create_metadata_object(c_s)
 
     def _create_metadata_object(self, c_s):
+        if not hasattr(c_s, "keys"):
+            raise ValueError("Metadata must be a YAML dictionary")
         keys = list(c_s.keys())
         engine = "Unknown"
 
@@ -305,7 +405,7 @@ class CollectionYamlLoader:
                 t = s[table]
                 tables.append(self.load_table(schema, table, t))
 
-        return Metadata(tables, engine)
+        return Metadata(tables, engine, dbname=collection)
 
     def load_table(self, schema, table, t):
         rowcount = int(t["rows"]) if "rows" in t else 0
@@ -342,36 +442,53 @@ class CollectionYamlLoader:
             schema,
             table,
             columns,
-            rowcount,
-            rows_exact,
-            row_privacy,
-            max_ids,
-            sample_max_ids,
-            clamp_counts,
-            clamp_columns,
-            use_dpsu,
-            censor_dims,
+            rowcount=rowcount,
+            rows_exact=rows_exact,
+            row_privacy=row_privacy,
+            max_ids=max_ids,
+            sample_max_ids=sample_max_ids,
+            clamp_counts=clamp_counts,
+            clamp_columns=clamp_columns,
+            use_dpsu=use_dpsu,
+            censor_dims=censor_dims,
         )
 
     def load_column(self, column, c):
+        lower = float(c["lower"]) if "lower" in c else None
+        upper = float(c["upper"]) if "upper" in c else None
         is_key = False if "private_id" not in c else bool(c["private_id"])
-        bounded = False if "bounded" not in c else bool(c["bounded"])
+        nullable = bool(c["nullable"]) if "nullable" in c else True
+        missing_value = c["missing_value"] if "missing_value" in c else None
+        sensitivity = c["sensitivity"] if "sensitivity" in c else None
 
         if c["type"] == "boolean":
-            return Boolean(column, is_key, bounded)
+            return Boolean(column, is_key=is_key, nullable=nullable, missing_value=missing_value)
         elif c["type"] == "datetime":
-            return DateTime(column, is_key, bounded)
+            return DateTime(column, is_key=is_key, nullable=nullable, missing_value=missing_value)
         elif c["type"] == "int":
-            minval = int(c["lower"]) if "lower" in c else None
-            maxval = int(c["upper"]) if "upper" in c else None
-            return Int(column, minval, maxval, is_key, bounded)
+            return Int(
+                column, 
+                lower=int(lower) if lower is not None else None,
+                upper=int(upper) if upper is not None else None,
+                is_key=is_key,
+                nullable=nullable,
+                missing_value=missing_value,
+                sensitivity=sensitivity
+            )
         elif c["type"] == "float":
-            minval = float(c["lower"]) if "lower" in c else None
-            maxval = float(c["upper"]) if "upper" in c else None
-            return Float(column, minval, maxval, is_key, bounded)
+            sensitivity = c["sensitivity"] if "sensitivity" in c else None
+            return Float(
+                column, 
+                lower=lower, 
+                upper=upper, 
+                is_key=is_key,
+                nullable=nullable,
+                missing_value=missing_value,
+                sensitivity=sensitivity
+            )
         elif c["type"] == "string":
             card = int(c["cardinality"]) if "cardinality" in c else 0
-            return String(column, card, is_key, bounded)
+            return String(column, card=card, is_key=is_key, nullable=nullable, missing_value=missing_value)
         else:
             raise ValueError("Unknown column type for column {0}: {1}".format(column, c))
 
@@ -417,14 +534,18 @@ class CollectionYamlLoader:
                     )
                 table[cname] = {}
                 column = table[cname]
-                if hasattr(c, "bounded") and c.bounded == True:
-                    column["bounded"] = c.bounded
                 if hasattr(c, "card"):
                     column["cardinality"] = c.card
-                if hasattr(c, "minval") and c.minval is not None:
-                    column["lower"] = c.minval
-                if hasattr(c, "maxval") and c.maxval is not None:
-                    column["upper"] = c.maxval
+                if hasattr(c, "lower") and c.lower is not None:
+                    column["lower"] = c.lower
+                if hasattr(c, "upper") and c.upper is not None:
+                    column["upper"] = c.upper
+                if hasattr(c, "nullable") and c.nullable is not None:
+                    column["nullable"] = c.nullable
+                if hasattr(c, "missing_value") and c.missing_value is not None:
+                    column["missing_value"] = c.missing_value
+                if hasattr(c, "sensitivity") and c.sensitivity is not None:
+                    column["sensitivity"] = c.sensitivity
                 if c.is_key is not None and c.is_key == True:
                     column["private_id"] = c.is_key
                 if type(c) is String:
